#include "settings.h"

Settings* Settings::Instance () 
{
	static Settings t;
	return &t;
}

Settings::Settings()
{
	profileName			= "default";

	registerCVar("fullscreen",						0, 0, 1, true, "enable fullscreen mode");
	registerCVar("headless",						0, 0, 1, true, "do not open gl context");
	registerCVar("fsX",							800, 1, 1000000, false, "fullscreen resolution X");
	registerCVar("fsY",							600, 1, 1000000, false, "fullscreen resolution Y");

	registerCVar("race",							0, 0, 1, false, "enable race simulation");
	registerCVar("testworld",						0, 0, 1, false, "enable race simulation");

	registerCVar("worldsizeX",						23, 1, 5000, false, "size of the world along axis X");
	registerCVar("worldsizeY",						13, 1, 5000, false, "size of the world along axis Y");
	registerCVar("worldwalls",						1, 0, 1, false, "enable walls around the world");

	registerCVar("killhalfdecreaseenergybypct",				1, 0, 100, false, "decrease energy by n percent when killhalfat triggers");

	registerCVar("energy",							400, 0, 1000000, false, "energy in the system by number of food cubes");
	registerCVar("mincritters",						10, 0, 1000, false, "minimum number of critters");

	registerCVar("insertcritterevery",					0, 0, 1000000, false, "inserts a random critter every n frames");
	registerCVar("retinasperrow",						20, 1, 1000, false, "number of vision retinas to stack per row onscreen");
	registerCVar("camerasensitivity",					20, 1, 1000, false, "sensitivity of the camera");

	registerCVar("colormode",						0, 0, 1, true, "colors genetically exact critters identically");
	registerCVar("exit_if_empty",						0, 0, 1, true, "exit simulation if there are no critters");
	registerCVar("autoload",						0, 0, 1, true, "autoload critters from ~/.critterding/load");


	registerCVar("critter_maxlifetime",					20000, 1, 10000000, false, "maximum number of frames a critter lives");
	registerCVar("critter_maxenergy",					5000, 1, 1000000, false, "maximum amount of energy a critter has");

	registerCVar("critter_startenergy",					3000, 1, 1000000, false, "energy a new critter starts with");
	registerCVar("critter_procinterval",					20, 1, 1000000, false, "minimum frames between procreations");
	registerCVar("critter_minenergyproc",					3000, 1, 1000000, false, "energy a critters needs to procreate");
	registerCVar("critter_sightrange",					70, 1, 1000000, false, "distance a critter can see (10 = 1 worldsize)");

	registerCVar("critter_retinasize",					16, 1, 1000, false, "size of a critters eye retina");
	registerCVar("critter_autosaveinterval",				0, 0, 1000000, false, "save critters every n seconds");
	registerCVar("critter_autoexchangeinterval",				0, 0, 1000000, false, "save critters every n seconds");
	registerCVar("critter_killhalfat",					120, 2, 1000000, false, "kill 50% of critters if population reaches n");
	registerCVar("critter_enableomnivores",					1, 0, 1, true, "enables critters to eat each other");
	registerCVar("critter_raycastvision",					0, 0, 1, true, "use raycast vision instead of opengl");
	
	registerCVar("food_maxlifetime",					10000, 1, 1000000, false, "maximum number of frames a food unit exists");
	registerCVar("food_maxenergy",						1500, 1, 1000000, false, "maximum amount of energy a food unit has");
	registerCVar("food_size",						200, 1, 1000000, false, "size of a food unit");

	registerCVar("pill_size",						40, 1, 1000000, false, "size of an excitatory agonist etc");
	registerCVar("pill_maxlifetime",					10000, 1, 1000000, false, "lifetime of a psychedelic");
	registerCVar("pill_energylevel",					10000, 1, 1000000, false, "dose per drug");
	registerCVar("pill_maxtrip",						100000, 1, 1000000, false, "max level of a drug (be careful & kind)");

	registerCVar("auto_inhib_antag_amt",					0, 0, 1000000, false, "docs");
	registerCVar("auto_inhib_antag_every",					0, 0, 10000000, false, "docs");
	registerCVar("auto_excit_antag_amt",					0, 0, 1000000, false, "docs");
	registerCVar("auto_excit_antag_every",					0, 0, 10000000, false, "docs");
	registerCVar("auto_excit_agon_amt",					0, 0, 1000000, false, "docs");
	registerCVar("auto_excit_agon_every",					0, 0, 10000000, false, "docs");
	registerCVar("auto_inhib_agon_amt",					0, 0, 1000000, false, "docs");
	registerCVar("auto_inhib_agon_every",					0, 0, 10000000, false, "docs");
	registerCVar("auto_potentiator_amt",					0, 0, 1000000, false, "docs");
	registerCVar("auto_potentiator_every",					0, 0, 10000000, false, "docs");
	registerCVar("autodecrement_autodrug_amts",				1, 0, 1, true, "docs");
	registerCVar("autodrug_autodecrement_amt",				0, 0, 1000000, false, "docs");


	registerCVar("body_maxmutations",					10, 1, 1000000, false, "maximum mutations on a body mutant");
	registerCVar("body_mutationrate",					10, 0, 100, false, "percentage of newborns that mutate bodies");

	registerCVar("body_maxbodyparts",					30, 0, 1000, false, "maximum body parts per critter");
	registerCVar("body_maxbodypartsatbuildtime",				6, 1, 1000, false, "maximum body parts for a new critter");

	registerCVar("body_minbodypartsize",					20, 1, 1000000, false, "minimum size of a critters body part");
	registerCVar("body_maxbodypartsize",					200, 1, 1000000, false, "maximum size of a critters body part");
	registerCVar("body_minheadsize",					30, 1, 1000000, false, "minimum size of a critters head");
	registerCVar("body_maxheadsize",					80, 1, 1000000, false, "maximum size of a critters head");

	registerCVar("body_percentmutateeffectaddbodypart",			3, 0, 100, false, "chance of adding a body part");
	registerCVar("body_percentmutateeffectremovebodypart",			3, 0, 100, false, "chance of removing a body part");
	registerCVar("body_percentmutateeffectresizebodypart",			3, 0, 100, false, "chance of resizing a body part");
	registerCVar("body_percentmutateeffectresizebodypart_slightly",		15, 0, 100, false, "chance of slightly resizing a body part");
	registerCVar("body_percentmutateeffectchangeconstraintlimits",		3, 0, 100, false, "chance of changing a joints motion limits");
	registerCVar("body_percentmutateeffectchangeconstraintlimits_slightly",	15, 0, 100, false, "chance of slightly changing a joints motion limits");
	registerCVar("body_percentmutateeffectchangeconstraintangles",		3, 0, 100, false, "chance of changing a joints position angles");
	registerCVar("body_percentmutateeffectchangeconstraintangles_slightly",	15, 0, 100, false, "chance of changing a joints position angles");
	registerCVar("body_percentmutateeffectchangeconstraintposition",	3, 0, 100, false, "chance of changing a joints position");
	registerCVar("body_percentmutateeffectchangeconstraintposition_slightly", 15, 0, 100, false, "chance of slightly changing a joints position");
	registerCVar("body_percentmutateeffectresizehead",			4, 0, 100, false, "chance of resizing a head");
	registerCVar("body_percentmutateeffectresizehead_slightly",		15, 0, 100, false, "chance of slightly resizing a head");
	registerCVar("body_percentmutateeffectrepositionhead",			3, 0, 100, false, "chance of repositioning head");

	registerCVar("brain_maxmutations",					10, 1, 1000000, false, "maximum mutations on a brain mutant");
	registerCVar("brain_mutationrate",					10, 0, 100, false, "percentage of newborns that mutate brains");

	registerCVar("brain_maxneurons",					1000, 1, 1000000, false, "maximum neurons per critter");
	registerCVar("brain_minsynapses",					1, 1, 1000000, false, "minimum synapses per neuron");
	registerCVar("brain_maxsynapses",					100, 1, 1000000, false, "maximum synapses per neuron");
	registerCVar("brain_minneuronsatbuildtime",				50, 1, 1000000, false, "minimum neurons for a new critter");
	registerCVar("brain_maxneuronsatbuildtime",				200, 1, 1000000, false, "maximum neurons for a new critter");

	registerCVar("brain_minsynapsesatbuildtime",				1, 1, 1000000, false, "minimum synapses for a new neuron");
	registerCVar("brain_maxsynapsesatbuildtime",				40, 1, 1000000, false, "maximum synapses for a new neuron of a new critter");

	registerCVar("brain_percentchanceinhibitoryneuron",			50, 0, 100, false, "percent chance a neuron is inhibotory");
	registerCVar("brain_mutate_percentchanceinhibitoryneuron",		0, 0, 1, true, "mutate this value");

	registerCVar("brain_percentchancemotorneuron",				50, 0, 100, false, "percent chance a neuron is a motor neuron");
	registerCVar("brain_mutate_percentchancemotorneuron",			0, 0, 1, true, "mutate this value");

	registerCVar("brain_percentchanceplasticneuron",			20, 0, 100, false, "percent chance a neuron has plastic synapses");
	registerCVar("brain_mutate_percentchanceplasticneuron",			0, 0, 1, true, "mutate this value");

	registerCVar("brain_minplasticitystrengthen",				100, 1, 1000000, false, "minimum weight by which plastic synapses strengthen");
	registerCVar("brain_maxplasticitystrengthen",				1000, 1, 1000000, false, "maximum weight by which plastic synapses strengthen");
	registerCVar("brain_minplasticityweaken",				1000, 1, 1000000, false, "minimum weight by which plastic synapses weaken");
	registerCVar("brain_maxplasticityweaken",				10000, 1, 1000000, false, "maximum weight by which plastic synapses weaken");
	registerCVar("brain_mutate_plasticityfactors",				0, 0, 1, true, "mutate min/max plasticity values");

	registerCVar("brain_minfiringthreshold",				2, 1, 1000000, false, "minimum firingthreshold of a neuron");
	registerCVar("brain_mutate_minfiringthreshold",				0, 0, 1, true, "mutate this value");

	registerCVar("brain_maxfiringthreshold",				10, 1, 1000000, false, "maximum firingthreshold of a neuron");
	registerCVar("brain_mutate_maxfiringthreshold",				0, 0, 1, true, "mutate this value");

	registerCVar("brain_maxdendridicbranches",				3, 1, 1000000, false, "maximum number of dendrites per neuron");
	registerCVar("brain_mutate_maxdendridicbranches",			0, 0, 1, true, "mutate this value");

	registerCVar("brain_percentchanceconsistentsynapses",			0, 0, 100, false, "percent chance a neurons synapses are all inhibitory or excitatory");
	registerCVar("brain_mutate_percentchanceconsistentsynapses",		0, 0, 1, true, "mutate this value");

	registerCVar("brain_percentchanceinhibitorysynapses",			50, 0, 100, false, "percent chance a synapse is inhibitory");
	registerCVar("brain_mutate_percentchanceinhibitorysynapses",		0, 0, 1, true, "mutate this value");

	registerCVar("brain_percentchancesensorysynapse",			20, 0, 100, false, "percent change a synapse connects to an input");
	registerCVar("brain_mutate_percentchancesensorysynapse",		0, 0, 1, true, "mutate this value");

	registerCVar("brain_percentmutateeffectaddneuron",			10, 0, 100, false, "chance of adding a neuron");
	registerCVar("brain_percentmutateeffectremoveneuron",			10, 0, 100, false, "chance of removing a neuron");
	registerCVar("brain_percentmutateeffectalterneuron",			20, 0, 100, false, "chance of altering a neuron");
	registerCVar("brain_percentmutateeffectaddsynapse",			29, 0, 100, false, "chance of adding a synapse");
	registerCVar("brain_percentmutateeffectremovesynapse",			29, 0, 100, false, "chance of removing a synapse");
	registerCVar("brain_mutate_mutateeffects",				0, 0, 1, true, "mutate mutation effects");

	registerCVar("brain_percentmutateeffectaltermutable",			2, 0, 100, false,"mutate value of a mutatable option");

	registerCVar("brain_costhavingneuron",					50, 0, 1000000, false, "cost of having a neuron (1/100000 energy)");
	registerCVar("brain_costfiringneuron",					10, 0, 1000000, false, "cost of firing a neuron");
	registerCVar("brain_costfiringmotorneuron",				100, 0, 1000000, false, "cost of firing a motor neuron");
	registerCVar("brain_costhavingsynapse",					1, 0, 1000000, false, "cost of having a synapse");

	registerCVar("brain_costhavingsynapse",					1, 0, 1000000, false, "cost of having a synapse");

	registerCVar("condensed_colourmotors",					1, 0, 1, true, "use 1 in 3 colour motors and force one colour for writes per frame");
	registerCVar("whitescreen",						0, 0, 1, true, "start psychic movie white not black (can save these soon)");
	registerCVar("autoload_movie",						0, 0, 1, true, "autoload movie from ~/.critterding/load/movie");
	registerCVar("anim_screen_every",					1, 1, 1000, false, "how many frames to pause between shared grid animation");
	registerCVar("show_scrn.pnm_every",					0, 0, 1000000, false, "flash screenshots on the shared grid, from ~/.critterding/screenshots/scrn.pnm");

	registerCVar("test_no_offspring",					0, 0, 1000000, false, "for simulating a lone critter, breeding does nothing but cost energy");


	info_critters = 0;
	info_food = 0;
	info_Graph_ad = 0.0f;    // flamoot
	info_Graph_avgsyn = 0;// flamoot
	info_Graph_avgneur= 0;// flamoot
	info_Graph_synvneur = 0.0f; // flamoot

	info_totalNeurons = 0;
	info_totalSynapses = 0;
	info_totalAdamDistance = 0;

	info_totalBodyparts = 0;
	info_totalWeight = 0;

	info_drugtotlev = 0;
	info_drugonelev = 0;
	info_drugtwolev = 0;
	info_drugthrlev = 0;
	info_drugfoulev = 0;
	info_drugfivlev = 0;
	
	createHelpInfo();
}

void Settings::registerCVar(const string& name, const unsigned int& defaultvalue, const unsigned int& min_val, const unsigned int& max_val, const string& comment)
{
	cerr << "hello" << endl;
	registerCVar( name, defaultvalue, min_val, max_val, false, comment );
}

void Settings::registerCVar(const string& name, const unsigned int& defaultvalue, const unsigned int& min_val, const unsigned int& max_val, bool loop, const string& comment)
{
	cvar* v = new cvar();

	v->int_val = defaultvalue;
	v->int_min = min_val;
	v->int_max = max_val;
	v->loop = loop;
	v->comment = comment;

	unregisterCVar(name);
	cvarlist[name] = v;

// 	cvarlist[name]->loop = true;
}

unsigned int Settings::getCVar(const string& name)
{
// 	return cvarlist[name]->int_val;
	
	cvarit = cvarlist.find(name);
	if ( cvarit != cvarlist.end() )
		return cvarit->second->int_val;
	else 
		cerr << "getCVar: no such key: " << name << endl;

	return 0;
}

const unsigned int* Settings::getCVarPtr(const string& name)
{
	return &cvarlist[name]->int_val;
}

bool Settings::setCVar(const string& name, const unsigned int& value)
{
	if ( value >= cvarlist[name]->int_min && value <= cvarlist[name]->int_max )
	{
		cvarlist[name]->int_val = value;
		return true;
	}
	else
	{
		cerr << name << " expects a value that is >=" << cvarlist[name]->int_min << " and <=" << cvarlist[name]->int_max << endl;
		return false;
	}
}

// FIXME private
void Settings::increaseCVar(const string& name, const unsigned int& value)
{
	unsigned int diff = cvarlist[name]->int_val + value;
	if ( diff >= cvarlist[name]->int_min && diff <= cvarlist[name]->int_max )
		cvarlist[name]->int_val += value;
	else if ( cvarlist[name]->loop )
	{
		cvarlist[name]->int_val = cvarlist[name]->int_min;
	}
}

// FIXME private
void Settings::decreaseCVar(const string& name, const unsigned int& value)
{
	unsigned int diff = cvarlist[name]->int_val - value;
	if ( diff >= cvarlist[name]->int_min && diff <= cvarlist[name]->int_max )
		cvarlist[name]->int_val -= value;
}

void Settings::increaseCVar(const string& name)
{
	unsigned int diff = cvarlist[name]->int_val + 1;
	if ( diff >= cvarlist[name]->int_min && diff <= cvarlist[name]->int_max )
		cvarlist[name]->int_val = diff;
	else if ( cvarlist[name]->loop )
	{
		cvarlist[name]->int_val = cvarlist[name]->int_min;
	}
	stringstream buf;
	buf << name << ": " << getCVar(name);
	Logbuffer::Instance()->add(buf);
}

void Settings::decreaseCVar(const string& name)
{
	int diff = -1 + cvarlist[name]->int_val;
	if ( diff >= (int)cvarlist[name]->int_min && diff <= (int)cvarlist[name]->int_max )
		cvarlist[name]->int_val = (unsigned int)diff;
	stringstream buf;
	buf << name << ": " << getCVar(name);
	Logbuffer::Instance()->add(buf);
}

void Settings::unregisterCVar(const string& name)
{
	cvarit = cvarlist.find(name);
	if ( cvarit != cvarlist.end() )
	{
		delete cvarit->second;
		cvarlist.erase(cvarit->first);
	}
}

bool Settings::isCVar(const string& name)
{
	cvarit = cvarlist.find(name);
	if ( cvarit != cvarlist.end() )
		return true;

	return false;
}

void Settings::createHelpInfo()
{
// 	helpinfo << "STARTUP OPTIONS" << endl << endl;
// 	helpinfo << "  Global Settings" << endl;

	helpinfo << endl << "  option                                                    [default] [range]      [comment]" << endl << endl;
	for( cvarit = cvarlist.begin(); cvarit != cvarlist.end(); cvarit++ )
	{
		stringstream buf("");
		buf << "  --" << cvarit->first;
// 		buf << " " << buf.str().size();
		
		for ( unsigned int i=buf.str().size(); i < 62; i++ )
			buf << " ";
		
		buf << cvarit->second->int_val;

		for ( unsigned int i=buf.str().size(); i < 72; i++ )
			buf << " ";

		buf << cvarit->second->int_min << "-" << cvarit->second->int_max;

		for ( unsigned int i=buf.str().size(); i < 85; i++ )
			buf << " ";

		buf << cvarit->second->comment;

		helpinfo << buf.str() << endl;
	}

	helpinfo << endl << " To save the default settings to a profile, press \"s\" in the simulation.  \n It will be saved to ./default and can be loaded by using \"--profile ./default\"" << endl ;
	helpinfo << endl << " Use F1 in the simulation for more information about keys." << endl;

}

void Settings::loadProfile(char* filename)
{
	cerr << "loading settins from '" << filename << "'" << endl;

	// nasty code for stripping dir names from the given profile name
	profileName = filename;
	string dirs("");
	dirs = parseH->Instance()->returnUntillStrip( "/", profileName );
	while ( !dirs.empty() )
	{
		dirs="";
		dirs = parseH->Instance()->returnUntillStrip( "/", profileName );
	}

	string content;
	if ( fileH.open( filename, content, 0 ) )
	{
		content.append("\n");
		string line = parseH->Instance()->returnUntillStrip( "\n", content );

		while ( !content.empty() )
		{
			// trim spaces
			while ( parseH->Instance()->beginMatchesStrip( " ", line ) ) {};

			// remove comment lines
			if ( parseH->Instance()->beginMatchesStrip( "#", line ) )
				line.clear();

			if ( !line.empty() )
			{
				string sw = parseH->Instance()->returnUntillStrip( " ", line );
				if ( !sw.empty() )
				{
					if ( isCVar(sw) )
					{
						while ( parseH->Instance()->beginMatchesStrip( " ", line ) ) {};
						if ( !line.empty() )
						{
							unsigned int value = atoi( line.c_str() );
							line.clear();

							if ( !setCVar(sw, value) )
								exit(1);
						}
						else
						{
							cerr << "Option expects an argument: " << sw << endl;
							exit(1);
						}

					}
					else
					{
						cerr << "Unknown option in profile: " << sw << endl;
						exit(1);
					}
				}
				else
				{
					cerr << "Option without an argument: " << line << endl;
					exit(1);
				}
			}
			line = parseH->Instance()->returnUntillStrip( "\n", content );
		}
	}
	else
	{
		cerr << "cannot open profile '" << filename << "'" << endl;
		exit(1);
	}
}

void Settings::saveProfile()
{
	stringstream buf;

	for( cvarit = cvarlist.begin(); cvarit != cvarlist.end(); cvarit++ )
		buf << cvarit->first << " " << cvarit->second->int_val << endl;

	fileH.save(profileName, buf.str());
}

void Settings::doCommandLineOptions(int argc, char *argv[])
{

	// first check if --help occurs, overrides the rest
	for (int i=1; i < argc; i++ )
	{
		string sw = argv[i];
		if ( sw == "--help" )
		{
			cout << helpinfo.str() << endl;
			exit(1);
	        }
	}

	// decode arguments
	int optind=1;
	while ((optind < argc) && (argv[optind][0]=='-'))
	{
		string sw = argv[optind];

		if ( sw=="--profile" )
		{
			if ( argv[++optind] )
			{
				loadProfile(argv[optind]);
			}
			else
			{
				cerr << "--profile expects a filename" << endl;
				exit(1);
			}
	        }

		else if ( parseH->Instance()->beginMatchesStrip( "--", sw ) )
		{
			if ( isCVar(sw) )
			{
				if ( argv[++optind] )
				{
					if ( !setCVar(sw, atoi( argv[optind] )) )
						exit(1);
				}
				else
				{
					cerr << "--" << sw << " expects an argument" << endl;
					exit(1);
				}
			}
			else
			{
				cerr << "Unknown commandline option: --" << sw << endl;
				cerr << "use --help for more information" << endl;
				exit(1);
			}
		}

		optind++;
	}
 
	if (optind<argc)
	{
		cerr << "Unknown commandline option: " << argv[optind] << endl;
		cerr << "use --help for more information" << endl;
		exit(1);
	}
	cout << endl << endl;
}

Settings::~Settings()
{
	for( cvarit = cvarlist.begin(); cvarit != cvarlist.end(); cvarit++ )
		delete cvarit->second;
}
